/*
 * Support for periodic interrupts (100 per second) and for getting
 * the current time from the RTC on Power Macintoshes.
 *
 * We use the decrementer register for our periodic interrupts.
 * On processors other than the 601, we calibrate the decrementer
 * frequency using the VIA timer 1.
 *
 * Paul Mackerras	August 1996.
 * Copyright (C) 1996 Paul Mackerras.
 *
 * Mostly not his code anymore; thanks for the names
 * Copyright (C) 1998 Teleworld, Inc. All Rights Reserved.
 */
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/init.h>

#include <asm/system.h>
#include <asm/ppcops.h>

/* This value is not really used. the following two variables are
 * overwritten later with valid values and not these stupid values
 * The following function does it
 */
#define TCD_CLOCK      (54)
#define MILLION        (1000000)
static unsigned decrementer_count = (TCD_CLOCK * MILLION) / HZ;
static unsigned counts_per_usec	  = TCD_CLOCK;
extern unsigned Xcounts_per_usec;
extern unsigned Xppc_tblo;

void
timer_clock_init(int cpuclock /* in MHz */)
{
	decrementer_count = (cpuclock * MILLION) / HZ;
	counts_per_usec   = cpuclock;
	Xcounts_per_usec   = counts_per_usec;
}

/*
 * This version of gettimeofday has microsecond resolution.
 */
void
do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;
	long pit;

	save_flags(flags);
	cli();
	*tv = xtime;
	pit = mfspr(SPRN_PIT);
	if (pit < 0) pit = 0;
	tv->tv_usec += (decrementer_count - pit) / counts_per_usec;
	if (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
	restore_flags(flags);
}

void
do_settimeofday(struct timeval *tv)
{
	unsigned long flags;
	unsigned int usec;
	long pit;

	save_flags(flags);
	cli();
	pit = mfspr(SPRN_PIT);
	if (pit < 0) pit = 0;
	usec = (decrementer_count - pit) / counts_per_usec;
	xtime.tv_sec = tv->tv_sec;
	xtime.tv_usec = tv->tv_usec - usec;
	if (xtime.tv_usec < 0) {
		xtime.tv_usec += 1000000;
		xtime.tv_sec--;
	}
	restore_flags(flags);
}

/*
 * timer_interrupt - gets called when the decrementer overflows,
 * with interrupts disabled.
 * We set it up to overflow again in 1/HZ seconds.
 */
void excprint(struct pt_regs *);
void prregs(struct pt_regs *);

void timer_interrupt(struct pt_regs * regs)
{
	mtspr(SPRN_TSR, 0x8000000);	/* acknowledge interrupt */
	do_timer(regs);
	//Xppc_tblo = mfspr(SPRN_TBLO);
}

void __init
time_init(void)
{
	int oiocr;

	xtime.tv_sec = 0; /* XXX */
	xtime.tv_usec = 0;

 	/* insure running clock-doubled */
 	oiocr = mfdcr(SPRN_IOCR);
 	oiocr |= 0x00004000;
 	mtdcr(SPRN_IOCR, oiocr);

	mtspr(SPRN_TCR, 0x04400000);	/* enable auto reload */
	mtspr(SPRN_TSR, 0xcc00000);	/* clear any lingering state */
 	mtspr(SPRN_PIT, decrementer_count);
	mtspr(SPRN_TBLO, 0);
}
