#define _AWACS_PCM_C_
/*
 * sound/awacs_pcm.c
 *
 * The low level driver for the Audio Waveform Amplifier & Converters (AWACs) Chip.
 */
	 
#include <linux/config.h>
#include "sound_config.h"

#if defined(CONFIG_AWACS) && defined(CONFIG_AUDIO)

#include <asm/dbdma.h>
#include <linux/delay.h>

#include "awacs.h"

#undef DEB
#define DEB(what)	/* (what) */
#undef DEB2
#define DEB2(what)	(what)

#define AWACS_PCM_INTRBITS (0x08)
/*
 * Sample buffer timer interrupt enable
 */

unsigned long	awacs_speed = 0;			/* sampling rate */
unsigned long	vspeed = 0;				/* virtual sampling rate */
short			awacs_channels = 1;		/* channels (1 or 2) */
unsigned char	awacs_bits = 16;			/* bits/sample (8 or 16) */
unsigned short	awacs_bitsok = AFMT_U8 | AFMT_S16_LE | AFMT_S16_BE;	/* mask of OK bits */
unsigned char	awacs_type = 0;			/* data type -- see above (ok bits) */
unsigned char	awacs_swap = 0;			/* swap bytes? */
unsigned long	awacs_count = 0;
int				awacs_busy = 0;
int				awacs_audiodev = 0;
int				open_mode = 0;

int
awacs_set_speed (int arg)
{
int	mask;

	DEB(printk("awacs_pcm.c: awacs_set_speed: arg = %d\n", arg));
	
	mask = ld_le32(AUD_CONT);
	DEB(printk("awacs_pcm.c: awacs_set_speed: AUD_CONT is first %#x\n", mask));
	
	if (arg != 0)
	{
		mask &= ~MASK_RATE;
		DEB(printk("awacs_pcm.c: awacs_set_speed: mask (pre) = %#x\n", mask));
	
		if (arg > 40000)
		{
			DEB(printk("awacs_pcm.c: awacs_set_speed: setting speed to %d (%#x).\n", 44100, RATE_44100));
			awacs_speed = 44100;
			mask |= RATE_44100;
		}
		else if (arg > 28000)
		{
			DEB(printk("awacs_pcm.c: awacs_set_speed: setting speed to %d (%#x).\n", 29400, RATE_29400));
			awacs_speed = 29400;
			mask |= RATE_29400;
		}
		else if (arg > 20000)
		{
			DEB(printk("awacs_pcm.c: awacs_set_speed: setting speed to %d (%#x).\n", 22050, RATE_22050));
			awacs_speed = 22050;
			mask |= RATE_22050;
		}
		else if (arg > 14000)
		{
			DEB(printk("awacs_pcm.c: awacs_set_speed: setting speed to %d (%#x).\n", 14700, RATE_14700));
			awacs_speed = 14700;
			mask |= RATE_14700;
		}
		else if (arg > 10800)
		{
			DEB(printk("awacs_pcm.c: awacs_set_speed: setting speed to %d (%#x).\n", 11025, RATE_11025));
			awacs_speed = 11025;
			mask |= RATE_11025;
		}
		else if (arg >= 8000)
		{
			DEB(printk("awacs_pcm.c: awacs_set_speed: setting speed to %d (%#x).\n", 8820, RATE_8820));
			awacs_speed = 8820;
			mask |= RATE_8820;
		}
		else
		{
			DEB(printk("awacs_pcm.c: awacs_set_speed: setting speed to %d (%#x).\n", 7350, RATE_7350));
			awacs_speed = 7350;
			mask |= RATE_7350;
		}

		/* put the hardware calls to actually _do_ the speed change here */
	
		st_le32(AUD_CONT, mask);
		eieio();
	}

	DEB(printk("awacs_pcm.c: awacs_set_speed: AUD_CONT is finally %#x\n", ld_le32(AUD_CONT)));
	
	if (arg == 8000)
	{
		/* hack for AUs */
		vspeed = 8000;
		return vspeed;
	}
	else if ((vspeed != 0) && (arg == 0))
		return vspeed;
	else
		vspeed = 0;
	
	switch(mask & MASK_RATE)
	{
		case RATE_44100:
			return (44100);

		case RATE_29400:
			return (29400);

		case RATE_22050:
			return (22050);

		case RATE_17640:
			return (17640);

		case RATE_14700:
			return (14700);

		case RATE_11025:
			return (11025);

		case RATE_8820:
			return (8820);

		case RATE_7350:
			return (7350);
	}

	return (arg);
	/* return (-EINVAL); */
}

short
awacs_set_channels (int arg)
{
	if ((arg != 1) && (arg != 2))
		return awacs_channels;

	if (arg != awacs_channels)
	{
		awacs_channels = arg;
	}

	return awacs_channels;
}

unsigned int
awacs_set_bits (int arg)
{
	if (arg == 0)
		return awacs_type;

	DEB(printk("awacs_pcm.c: awacs_set_bits(int arg = %d)\n", arg));
	
	if ((arg & awacs_bitsok) != arg)
		return awacs_type;
	
	if (arg != awacs_type)
	{
		switch (arg)
		{
			case AFMT_S16_LE:	/* signed 16-bit little-endian */
				awacs_bits = 16;
				awacs_swap = 0;
				awacs_type = AFMT_S16_LE;
				break;
			
			case AFMT_S16_BE:	/* signed 16-bit big-endian */
				awacs_bits = 16;
				awacs_swap = 1;
				awacs_type = AFMT_S16_BE;
				break;
				
			case AFMT_U8:		/* unsigned 8-bit */
				awacs_bits = 8;
				awacs_swap = 1;
				awacs_type = AFMT_U8;
				break;
			
			default:
				return (awacs_type);
		}
		
		awacs_set_swap(awacs_swap);
	}

	return awacs_bits;
}

int
awacs_set_swap (int arg)
{
	out_le32(AUD_BYTESWAP, arg);
	return (awacs_swap = in_le32(AUD_BYTESWAP));
}

int
awacs_audio_ioctl (int dev, unsigned int cmd, caddr_t arg, int local)
{
	int val;

	DEB(printk ("awacs_pcm.c: static int awacs_audio_ioctl(unsigned int cmd = %X, unsigned int arg = %p)\n", cmd, arg));

	switch (cmd)
	{
		case SOUND_PCM_WRITE_RATE:
			if (local)
				return awacs_set_speed ((int) arg);
			get_user(val, (int *) arg);
			return snd_ioctl_return ((int *) arg, awacs_set_speed (val));
			break;

		case SOUND_PCM_READ_RATE:
			if (local)
				return awacs_speed;
			return snd_ioctl_return ((int *) arg, awacs_speed);
			break;

		case SNDCTL_DSP_STEREO:
			if (local)
				return awacs_set_channels ((int) arg + 1) - 1;
			get_user(val, (int *) arg);
			return snd_ioctl_return ((int *) arg, awacs_set_channels (val + 1) - 1);
			break;

		case SOUND_PCM_WRITE_CHANNELS:
			if (local)
				return awacs_set_channels ((int) arg);
			get_user(val, (int *) arg);
			return snd_ioctl_return ((int *) arg, awacs_set_channels (val));
			break;

		case SOUND_PCM_READ_CHANNELS:
			if (local)
				return awacs_channels;
			return snd_ioctl_return ((int *) arg, awacs_channels);
			break;

		case SNDCTL_DSP_SETFMT:
			DEB(printk("SNDCTL_DSP_SETFMT: local = %d\n", local));
			if (local)
				return awacs_set_bits ((int) arg);
			get_user(val, (int *)arg);
			DEB(printk("SNDCTL_DSP_SETFMT: val = %d\n", (local? -1 : val)));
			return snd_ioctl_return ((int *) arg, awacs_set_bits (val));
			break;

		case SOUND_PCM_READ_BITS:
			if (local)
				return awacs_bits;
			return snd_ioctl_return ((int *) arg, awacs_bits);

		default:
			return -(EINVAL);
	}

	return -(EINVAL);
}

void
awacs_audio_reset (int dev)
{
DEB (printk ("awacs_pcm.c: static void awacs_audio_reset(void)\n"));

	/* some resetting thing here */
	awacs_busy = 0;
	out_le32(AUD_OUT_CHCONT, RUN << 16);	/* stop tx dma */
	out_le32(AUD_IN_CHCONT, RUN << 16);	/* stop rx dma */
}

int
awacs_audio_open (int dev, int mode)
{
int				err;

	DEB (printk ("awacs_pcm.c: static int awacs_audio_open(int mode = %X)\n", mode));

	awacs_count = 0;
	open_mode = mode;

	return 0;
}

void
awacs_audio_close (int dev)
{
unsigned long	flags;

	DEB (printk ("awacs_pcm.c: static void awacs_audio_close(void)\n"));

	save_flags (flags);
	cli ();

	awacs_audio_reset (dev);

	awacs_busy = 0;
	restore_flags (flags);
}

#define AWACS_BUFSIZE		262144			/* 256k = 64k system buffer * 2 bytes per channel per sample * 2 channels */
unsigned char awacs_buffer[AWACS_BUFSIZE];	

void
awacs_audio_output_block (int dev, unsigned long buf, int count,
			int intrflag, int restart_dma)
{
struct dbdma_cmd	*dcmd, *dp;
static struct dbdma_cmd	dcmd_space[AWACS_BUFSIZE / MAX_DMACOUNT + 3];
unsigned char *output_buffer, *tmp_buf;

	DEB (printk ("awacs_pcm.c: awacs_audio_output_block(%lx, %d)\n",
		     buf, count));
	
	awacs_busy = 1;

	DEB(printk("data prewrite = 0x[%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x]\n", 
			output_buffer[0], output_buffer[1], output_buffer[2], output_buffer[3], output_buffer[4], 
			output_buffer[5], output_buffer[6], output_buffer[7], output_buffer[8], output_buffer[9]));
	
	output_buffer = awacs_convert(phys_to_virt(buf), &count);
	
	dcmd = (struct dbdma_cmd *) DBDMA_ALIGN(dcmd_space);
	memset(dcmd, 0, 5 * sizeof(struct dbdma_cmd));
	
	st_le32(AUD_OUT_CHCONT, (RUN) << 16);	/* stop the channel */
	eieio();

	tmp_buf = (char *) output_buffer;
	for (dp = dcmd; count > MAX_DMACOUNT; ++dp) {
		init_dma(dp, OUTPUT_MORE, tmp_buf, MAX_DMACOUNT);
		tmp_buf += MAX_DMACOUNT;
		count -= MAX_DMACOUNT;
	}
	init_dma(dp, OUTPUT_LAST + INTR_ALWAYS, tmp_buf, count);
	init_dma(dp + 1, DBDMA_STOP, 0, 0);
	st_le32(AUD_OUT_CMDPTR, virt_to_phys(dcmd));
	eieio();
	st_le32(AUD_OUT_CHCONT, ((RUN) << 16) | (RUN));
	eieio();

	DEB(printk("data postwrite = 0x[%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x]\n", 
			output_buffer[0], output_buffer[1], output_buffer[2], output_buffer[3], output_buffer[4], 
			output_buffer[5], output_buffer[6], output_buffer[7], output_buffer[8], output_buffer[9]));
	
	DEB(printk("ld_le32(AUD_OUT_CHCONT) = %x\n", ld_le32(AUD_OUT_CHCONT)));
	DEB(printk("ld_le32(AUD_OUT_CHSTAT) = %x\n", ld_le32(AUD_OUT_CHSTAT)));
	
	DEB(printk("awacs_audio_output_block: Done.\n"));
}

unsigned char *
awacs_convert(char *buf, int *len)
{
int			i, j;

	if (awacs_bits == 16 && awacs_channels == 2)
	{
		/* memcpy(awacs_buffer, buf, *len); */
		return ((unsigned char *) buf);
	}
	
	else if (awacs_bits == 16 && awacs_channels == 1)
	{
		*len <<= 1;	/* *2 */
		for (i = j = 0; i < *len; i += 4, j += 2)
		{
			/* make right channel = left channel */
		
			awacs_buffer[i] = awacs_buffer[i+2] = buf[j];
			awacs_buffer[i+1] = awacs_buffer[i+3] = buf[j+1];
		}

		return (awacs_buffer);
	}
	
	else if (awacs_bits == 8 && awacs_channels == 2)
	{
		*len <<= 1;	/* *2 */
		for (i = j = 0; i < *len; i += 4, j += 2)
		{
			/* make low-byte = 0 (i.e., 8-bit - 16-bit conversion?) */
		
			awacs_buffer[i] = awacs_buffer[i+2] = 0;
			awacs_buffer[i+1] = buf[j] ^ 0x80;
			awacs_buffer[i+3] = buf[j+1] ^ 0x80;
		}

		return (awacs_buffer);
	}
	
	else if (awacs_bits == 8 && awacs_channels == 1)
	{
		*len <<= 2;	/* *4 */
		
		if (vspeed == 8000)
		{
			DEB(printk("playing 8kHz sample...\n"));
			
			/* convert sample from 8000Hz -> 8820Hz -- hack for AUs */
			for (i = j = 0; i < *len; i += 4, ++j)
			{
				/* both of the above */
				
				awacs_buffer[i] = awacs_buffer[i+2] = 0;
				awacs_buffer[i+1] = awacs_buffer[i+3] = buf[j] ^ 0x80;
				
				/* if ((j % 10) || (j % 441))	*/
				if (!(j & 0xf))		/* j % 16 */
				{
					i += 4;
					awacs_buffer[i] = awacs_buffer[i+2] = 0;
					awacs_buffer[i+1] = awacs_buffer[i+3] = (buf[j] ^ 0x80);
					/* awacs_buffer[i+1] = awacs_buffer[i+3] = (((buf[j] ^ 0x80) + (buf[j+1] ^ 0x80)) >> 1);	*/
					/* awacs_buffer[i+1] = awacs_buffer[i+3] = (((buf[j] + buf[j+1]) ^ 0x80) >> 1);	*/
				}
			}
		}
		
		else
		{
			DEB(printk("playing %dHz (%dHz) sample...\n", awacs_speed, vspeed));
			
			for (i = j = 0; i < *len; i += 4, ++j)
			{
				awacs_buffer[i] = awacs_buffer[i+2] = 0;
				awacs_buffer[i+1] = awacs_buffer[i+3] = buf[j] ^ 0x80;
			}
		}

		return (awacs_buffer);
	}

	return (awacs_buffer);
}

unsigned char input_buffer[AWACS_BUFSIZE];
unsigned char *system_input_buffer;
int system_input_length;

void
awacs_audio_start_input (int dev, unsigned long system_buffer, int count,
			 int intrflag, int restart_dma)
{
	struct dbdma_cmd	*dcmd, *dp;
	static struct dbdma_cmd	dcmd_space[AWACS_BUFSIZE / MAX_DMACOUNT + 3];
	int	i, len;
	unsigned char *buf;
	
	system_input_buffer = (char *) phys_to_virt(system_buffer);
	system_input_length = count;
	
	DEB(printk("system_buffer = %#x ; count = %d\n", system_buffer, count));
	DEB(printk("system_input_buffer = %#x ; system_input_length = %d\n", system_input_buffer, system_input_length));
	
	for (i = 0; i < AWACS_BUFSIZE; i += 4)
	{
		memset(input_buffer+i, 0xc0, 2);
		memset(input_buffer+i+2, 0x3f, 2);
	}
	
	DEB(printk("data preread = 0x[%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x]\n", 
			input_buffer[0], input_buffer[1], input_buffer[2], input_buffer[3], input_buffer[4], 
			input_buffer[5], input_buffer[6], input_buffer[7], input_buffer[8], input_buffer[9]));
	
	dcmd = (struct dbdma_cmd *) DBDMA_ALIGN(dcmd_space);
	memset(dcmd, 0, 2 * sizeof(struct dbdma_cmd));
	
	st_le32(AUD_IN_CHCONT, (RUN|PAUSE|FLUSH|WAKE) << 16);	/* stop the channel */
	eieio();
	buf = (char *) input_buffer;
	for (dp = dcmd; len > MAX_DMACOUNT; ++dp) {
		init_dma_rx(dp, INPUT_MORE, buf, MAX_DMACOUNT);
		buf += MAX_DMACOUNT;
		len -= MAX_DMACOUNT;
	}
	init_dma_rx(dp++, INPUT_LAST + INTR_ALWAYS, buf, len);
	init_dma_rx(dp++, DBDMA_STOP, 0, 0);
	init_dma_rx(dp++, DBDMA_NOP + BR_ALWAYS, 0, 0);
	st_le32(AUD_IN_CMDPTR, virt_to_bus(dcmd));
	eieio();
	st_le32(AUD_IN_CHCONT, ((RUN) << 16) | (RUN));
	eieio();
	
	DEB(printk("data postread = 0x[%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x]\n", 
			input_buffer[0], input_buffer[1], input_buffer[2], input_buffer[3], input_buffer[4], 
			input_buffer[5], input_buffer[6], input_buffer[7], input_buffer[8], input_buffer[9]));
	
	return;
}

int
awacs_audio_prepare_for_input (int dev, int bsize, int bcount)
{
	return 0;
}

int
awacs_audio_prepare_for_output (int dev, int bsize, int bcount)
{
	return 0;
}

void
awacs_halt_input (int dev)
{
	st_le32(AUD_IN_CHCONT, (RUN) << 16);	/* stop the channel */
	eieio();
}

void
awacs_halt_output (int dev)
{
	st_le32(AUD_OUT_CHCONT, (RUN) << 16);	/* stop the channel */
	eieio();
}

static struct audio_driver awacs_audio_driver =
{
	awacs_audio_open,
	awacs_audio_close,
	awacs_audio_output_block,
	awacs_audio_start_input,
	awacs_audio_ioctl,
	awacs_audio_prepare_for_input,
	awacs_audio_prepare_for_output,
	awacs_audio_reset,
	awacs_audio_reset,
	NULL,
	NULL,
	awacs_halt_input,
	awacs_halt_output,
	NULL,
	NULL,
	NULL,
	NULL
};

static struct audio_operations awacs_audio_operations =
{
	"AWACS Sound Chip",
	DMA_PSEUDO_AUTOMODE | DMA_DUPLEX,
	AFMT_U8 | AFMT_S16_LE | AFMT_S16_BE,
	NULL,
	&awacs_audio_driver
};

void
awacs_pcm_init (struct address_info *hw_config)
{
	DEB (printk ("awacs_pcm.c: long awacs_pcm_init()\n"));

	awacs_bitsok = AFMT_U8 | AFMT_S16_LE | AFMT_S16_BE;
	if (num_audiodevs < MAX_AUDIO_DEV)
	{
		audio_devs[awacs_audiodev = num_audiodevs++] = &awacs_audio_operations;
		audio_devs[awacs_audiodev]->dmachan1 = hw_config->dma;
		audio_devs[awacs_audiodev]->dmachan2 = hw_config->dma2;
		audio_devs[awacs_audiodev]->buffsize = DSP_BUFFSIZE;
		audio_devs[awacs_audiodev]->min_fragment = DSP_BUFFSIZE;
	}
	else
		printk ("AWACS: Too many PCM devices available\n");

	eieio();
	st_le32(AUD_CONT, 0x00004011);		/* main control, 44.1kHz, input & output on subframe 1 */
	eieio();
	st_le32(AUD_IN_INTRSEL, 0);
	st_le32(AUD_IN_BRSEL, 0);
	st_le32(AUD_IN_WAITSEL, 0);
	
	awacs_set_bits(AFMT_S16_LE);
	
	awacs_set_speed (DSP_DEFAULT_SPEED);
	
	request_irq(8, awacs_txdma_intr, 0, "AWACS-txdma", NULL);
	request_irq(9, awacs_rxdma_intr, 0, "AWACS-rxdma", NULL);
}

void
awacs_pcm_interrupt (unsigned char status, int cause)
{
	if (cause == 1)		/*
						 * PCM buffer done
						 */
	{
	}
}

void
awacs_txdma_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	if (awacs_busy) {
		awacs_busy = 0;
		DMAbuf_outputintr (awacs_audiodev, 1);
	}
}

void
awacs_rxdma_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	DEB(printk("awacs_rxdma_intr: dma input interrupt triggered.\n"));
	DEB(printk("data postread = 0x[%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x][%.2x]\n", 
			input_buffer[0], input_buffer[1], input_buffer[2], input_buffer[3], input_buffer[4], 
			input_buffer[5], input_buffer[6], input_buffer[7], input_buffer[8], input_buffer[9]));
	DEB(printk("system_input_buffer = %#x ; system_input_length = %d\n", system_input_buffer, system_input_length));
	
	if (system_input_length != 0) {
		memcpy(system_input_buffer, input_buffer, system_input_length);
		DMAbuf_inputintr (awacs_audiodev);
	}
}

#endif
