
//#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>

#if __KERNEL__
extern void ksprintf(char *buf,const char *fmt, ...);
extern void kvsprintf(char *buf, const char *cfmt, va_list ap);
#else
typedef unsigned long ulong;
#endif


typedef struct time {
	ulong hi;
	ulong lo;	
} TimeStamp;

static inline void Sync( void )
{    
    asm ( "sync" : : );
}

static inline ulong GetTBRLo( void )
{
    register ulong time;

    asm volatile( "mfspr %0,0x3CD" : "=r" (time) : );
	
    return time;
}

static inline ulong GetTBRHi( void )
{
    register ulong time;
    
    asm volatile( "mfspr %0,0x3CC" : "=r" (time) : );
	
    return time;
}


static inline void SubtractTime( TimeStamp *start, TimeStamp *stop, TimeStamp *result )
{
    
    asm( "subfc %0,%2,%4; subfe %1,%3,%5" 
	 : "=&r" (result->lo), "=&r" (result->hi) 
	 : "r" (start->lo), "r" (start->hi), "r" (stop->lo), "r" (stop->hi) );
}


static inline void GetTime( TimeStamp *timestamp )
{
    register ulong hi,lo;
	
    Sync();
    
    do
    {
	hi = GetTBRHi();
	lo = GetTBRLo();
    } while( hi != GetTBRHi() );

    timestamp->hi = hi;    
    timestamp->lo = lo;
}

static inline void ElapsedTime( TimeStamp *starttime, TimeStamp *result )
{            
    TimeStamp now;

    GetTime( &now );

    SubtractTime( starttime, &now, result );        
}

#define kMemLogBuffer     0xA0000000
#define kMemLogEntryCount 0x20000
#define kMemLogEntrySize  0x80

static inline int GetNextIndex(void)
{
    register ulong index;
    register ulong *memLogIndexPtr = (ulong*)kMemLogBuffer;

    // do an atomic increment

    asm( "loop:   lwarx  %0,0,%1; "
		  "addi   0,%0,1; "
		  "stwcx. 0,0,%1; "
		  "bne    loop    "
	 : "=&r" (index)
	 : "r" (memLogIndexPtr)
	 : "cr0","r0" );

    return index;
}


#if defined(DEBUG) && DEBUG

static inline void PostMemLog( char * format, ... )
{ 
    va_list ap;
    
    TimeStamp time;

    ulong index = GetNextIndex();

    char *memLogBuffer = (char*)kMemLogBuffer;    
    
    ulong* entry = (ulong*)&(memLogBuffer[ ( index % kMemLogEntryCount ) * kMemLogEntrySize ]);

    GetTime( &time );

    *entry++ = index + 1;
    
    *entry++ = time.lo;

    va_start( ap, format );
    
#if __KERNEL__
    kvsprintf( (char*)entry, format, ap );   
#else
    vsprintf( (char*)entry, format, ap );   
#endif
    va_end( ap );
}

#if !__KERNEL__

static inline void ResetMemLog( )
{
    *(ulong*)kMemLogBuffer = 0;
}

static inline void PrintMemLog( )
{
    ulong i;
    ulong index, time, lasttime = 0;  

    char *memLogBuffer = (char*)kMemLogBuffer;    

    PostMemLog( "Start PrintMemLog" );
    
    for( i = 0; i < kMemLogEntryCount; ++i )
    {	
	if( i >= *(ulong*)kMemLogBuffer )
	{
	    break;
	}
	index = *(ulong*)(memLogBuffer);

	time  = *(ulong*)(memLogBuffer + 4);

	printf( "%8lu %10lu %s\n", index, (time - lasttime)/54, (char*)memLogBuffer + 8 );

	lasttime = time;	

	memLogBuffer += kMemLogEntrySize;
    }
    ResetMemLog();
}


static inline void DumpMemLog( )
{
    ulong i;

    char *memLogBuffer = (char*)kMemLogBuffer;    
    
    PostMemLog( "Start DumpMemLog" );

    for( i = 0; i < kMemLogEntryCount; ++i )
    {	
	if( i >= *(ulong*)kMemLogBuffer ) 
	{
	    break;
	}
	
	printf( "%s",  memLogBuffer + 8 );

	memLogBuffer += kMemLogEntrySize;
    }

    ResetMemLog();
}


static inline void TestMemLog( )
{
    
    int i;

    TimeStamp starttime;

    GetTime( &starttime );

    for( i = 0; i < 60; ++i )
    {
	
	TimeStamp stoptime;
	    
	TimeStamp elapsed;

	GetTime( &stoptime );
	
	SubtractTime( &starttime, &stoptime, &elapsed );

	starttime = stoptime;

	PostMemLog( "test %4d at time: %08X %08X", i, elapsed.hi, elapsed.lo );
	
	sleep( 1 );	
    }
    PrintMemLog();
}

#endif //!__KERNEL__
#endif
